import logging, os
from pathlib import Path, PurePath

from PyPDF2 import PdfFileWriter, PdfFileReader, pdf
from pdfrw import PdfReader, PdfWriter, PageMerge
from bookleter import shuffle


class Book():
    def __init__(
        self,
        input_file_path,
        start_page_number,
        end_page_number,
        direction,
        crop,
        ):

        self._validate_inputs(
            input_file_path,
            start_page_number,
            end_page_number,
            crop
        )

        current_path = Path.cwd()
        self.input_file_path = PurePath.joinpath(current_path, input_file_path)
        self.start_page_number = int(start_page_number)
        self.end_page_number = int(end_page_number)
        self.direction = direction
        self.crop = crop

        logging.basicConfig(level=logging.NOTSET)

        self.original_pdf_name = self.input_file_path.name
        self.original_pdf_path = str(self.input_file_path)
        self.final_pdf_name = self.original_pdf_path.replace(".pdf", "_print_this.pdf")
        self.test_pdf_name = self.final_pdf_name.replace(".pdf", "_for_test.pdf")

    def make_booklet(self):

        page_numbers = list(range(self.start_page_number, self.end_page_number + 1))

        self.end_page_number = (self.end_page_number - self.start_page_number) + 1
        self.start_page_number = 1

        if self.end_page_number % 8 == 0:
            correct_pages_count = self.end_page_number
        else:
            correct_pages_count = ((((self.end_page_number - self.start_page_number) + 1) // 8) + 1) * 8
        blank_pages_count = correct_pages_count - self.end_page_number

        inputpdf = PdfFileReader(open(self.original_pdf_path, "rb"))
        final = PdfFileWriter()
        test_file = PdfFileWriter()

        print_order = shuffle.foop(correct_pages_count)
        test_print_order = shuffle.foop(8)

        for blank in range(blank_pages_count):
            page_numbers.append(-1)
        if self.direction == "rtl":
            page_numbers.reverse()
            test_print_order.reverse()

        for number in print_order:
            pg_number = page_numbers[number - 1]
            if pg_number == -1:
                # it's a blank page
                page = pdf.PageObject.createBlankPage(pdf=inputpdf)
                page.cropBox.lowerLeft = tuple([a + b for a, b in zip(page.cropBox.lowerLeft, (int(self.crop['left']), int(self.crop['bottom'])))])
                page.cropBox.upperRight = tuple([a - b for a, b in zip(page.cropBox.upperRight, (int(self.crop['right']), int(self.crop['top'])))])
                final.addPage(page)
            else:
                # it's a page from the original pdf
                page = inputpdf.getPage(pg_number - 1)
                # do the crops
                page.cropBox.lowerLeft = tuple([a + b for a, b in zip(page.cropBox.lowerLeft, (int(self.crop['left']), int(self.crop['bottom'])))])
                page.cropBox.upperRight = tuple([a - b for a, b in zip(page.cropBox.upperRight, (int(self.crop['right']), int(self.crop['top'])))])
                final.addPage(page)

        # if the pdf is smaller than 8 pages create no test file
        if self._get_pdf_pages_count(self.original_pdf_path) >= 8:
            for pg_number in test_print_order:
                test_file.addPage(inputpdf.getPage(pg_number - 1))
            with open(self.test_pdf_name, "wb") as output_stream:
                test_file.write(output_stream)

        with open(self.final_pdf_name, "wb") as output_stream:
            final.write(output_stream)

        self._make_four_in_one_pdf(self.final_pdf_name)
        self._make_four_in_one_pdf(self.test_pdf_name)

        # cleanup
        os.remove(self.final_pdf_name)
        os.remove(self.test_pdf_name)

    def _make_four_in_one_pdf(self, inputpdf):
        outputpdf = inputpdf.replace('.pdf', '_4in1.pdf')
        pages = PdfReader(inputpdf).pages
        writer = PdfWriter(outputpdf)

        # can't use a blank page in pdfrw lib
        # because of None existent page contents
        # so for every blank page we clone the page contents
        # from a non empty page
        non_empty_page_contents = None
        for page in pages:
            if page.Contents:
                non_empty_page_contents = page.Contents
        for page in pages:
            if not page.Contents:
                page.Contents = non_empty_page_contents

        for index in range(0, len(pages), 4):
            page = self._get_four_pages_as_one(pages[index:index + 4])
            writer.addpage(page)
        writer.write()


    def _get_four_pages_as_one(self, srcpages):
        scale = 0.5
        srcpages = PageMerge() + srcpages
        x_increment, y_increment = (scale * i for i in srcpages.xobj_box[2:])
        for i, page in enumerate(srcpages):
            page.scale(scale)
            page.x = x_increment if i & 1 else 0
            page.y = 0 if i & 2 else y_increment
        return srcpages.render()

    def _get_pdf_pages_count(self, input_file_path):
        pdf = PdfFileReader(open(input_file_path, "rb"))
        return pdf.getNumPages()

    def _validate_inputs(
            self,
            input_file_path,
            start_page_number,
            end_page_number,
            crop
        ):
        if input_file_path == "":
            raise ValueError('Please add a pdf file')

        if "" in (start_page_number, end_page_number):
            raise ValueError('Please enter start and end page numbers')
        else:
            try:
                int(start_page_number)
                int(end_page_number)
            except:
                raise ValueError('Start and end page value must be a number')

        if int(end_page_number) > self._get_pdf_pages_count(input_file_path):
            raise ValueError('End page number out of range\nYour book has only {} pages'.format(self._get_pdf_pages_count(input_file_path)))


        crop_values = ["" for key in crop.keys() if key == crop[key]]
        if "" in crop_values:
            raise ValueError('Please enter all the crop values')
        else:
            for val in crop.values():
                try:
                    int(val)
                except:
                    raise ValueError('Crop value must be a number')

    def __str__(self):
        return str({
            "pdf_file_path": self.input_file_path,
            "start_page": self.start_page_number,
            "end_page": self.end_page_number,
            "crop": self.crop,
            "book_direction": self.direction
        })



import pathlib, webbrowser
import tkinter as tk
from tkinter import filedialog, messagebox, Menu, Label, Toplevel
import pygubu
from bookleter import Booklet
from bookleter.__version__ import __version__

class Application:
    def __init__(self, master=None):
        self.master = master
        
        self.builder = builder = pygubu.Builder()
        current_path = pathlib.Path(__file__).parent.absolute()
        tkui_path = current_path / 'bookleter.ui'
        builder.add_from_file(tkui_path)
        self.mainwindow = builder.get_object('mainwindow', master)
        builder.connect_callbacks(self)

        menu = Menu(self.master)
        self.master.config(menu=menu)

        infoMenu = Menu(menu)
        menu.add_cascade(label="Hey", menu=infoMenu)
        infoMenu.add_command(label="About", command=self.show_about)

        callbacks = {
            'on_file_path_button_clicked': self.on_file_path_button_clicked,
            'on_make_booklet_button_clicked': self.on_make_booklet_button_clicked
            }

        builder.connect_callbacks(callbacks)

        combo = self.builder.get_object('book_direction_combobox')
        combo.bind("<<ComboboxSelected>>", self.on_book_direction_change)

        self.file_path = ""
        self.start_page_number = ""
        self.end_page_number = ""

        self.crop = {
            'left': "",
            'top': "",
            'right': "",
            'bottom': ""
        }

        self.builder.get_object('crop_left').insert(0, 'left')
        self.builder.get_object('crop_top').insert(0, 'top')
        self.builder.get_object('crop_right').insert(0, 'right')
        self.builder.get_object('crop_bottom').insert(0, 'bottom')

        for entry_id in ["crop_left", "crop_top", "crop_right", "crop_bottom"]:
            self.set_placeholder(entry_id)

        self.book_direction_index = 0


    def set_placeholder(self, entry_id):
        entry = self.builder.get_object(entry_id)
        entry.bind("<FocusIn>", lambda event, arg={"entry": entry, "placeholder": entry.get()}: self.foc_in(event, arg))
        entry.bind("<FocusOut>", lambda event, arg={"entry": entry, "placeholder": entry.get()}: self.foc_out(event, arg))

    def put_placeholder(self, entry, placeholder):
        entry.insert(0, placeholder)

    def foc_in(self, *args):
        try:
            int(args[1]['entry'].get())
        except Exception:
            args[1]['entry'].delete('0', 'end')

    def foc_out(self, *args):
        if not args[1]['entry'].get():
            self.put_placeholder(args[1]['entry'], args[1]['placeholder'])


    def on_file_path_button_clicked(self):
        self.file_path = file_path = filedialog.askopenfilename(title="Select Pdf File")
        if file_path:
            self.builder.get_object('file_name_label').config(text=pathlib.Path(self.file_path).name)

    def on_make_booklet_button_clicked(self):
        self.start_page_number = self.builder.get_object('start_page_input').get()
        self.end_page_number = self.builder.get_object('end_page_input').get()
        self.crop["left"] = self.builder.get_object('crop_left').get()
        self.crop["top"] = self.builder.get_object('crop_top').get()
        self.crop["right"] = self.builder.get_object('crop_right').get()
        self.crop["bottom"] = self.builder.get_object('crop_bottom').get()

        direction_options = ["rtl", "ltr"]

        # make booklet
        try :
            new_book = Booklet.Book(
                self.file_path,
                self.start_page_number,
                self.end_page_number,
                direction_options[self.book_direction_index],
                self.crop
            )
            new_book.make_booklet()
        except IndexError:
            messagebox.showerror("", "Page number out of range")
            return
        except Exception as e:
            messagebox.showerror("", e)
            return
        
        messagebox.showinfo("", "Your Booklet Is Ready!")
    
    def on_book_direction_change(self, event):
        combo = self.builder.get_object('book_direction_combobox')
        self.book_direction_index = combo.current()

    def show_about(self):
        ABOUT_TEXT = "The Bookleter {} ".format(__version__)
        toplevel = Toplevel(borderwidth=50)
        
        about_label = Label(toplevel, text=ABOUT_TEXT, height=0, width=0)
        about_label.pack()
        
        github_link = Label(toplevel, text="Bookleter on Github", fg="blue", cursor="hand2", pady=5)
        github_link.pack()
        github_link.bind("<Button-1>", lambda e: self.callback("http://github.com/reinenichts/bookleter"))

        virgool_link = Label(toplevel, text="About Booklets", fg="blue", cursor="hand2", pady=5)
        virgool_link.pack()
        virgool_link.bind("<Button-1>", lambda e: self.callback("https://virgool.io/@mohsenbarzegar/bookleter-nkkuh18xnbyk"))

        new_version_link = Label(toplevel, text="New Version", fg="blue", cursor="hand2", pady=5)
        new_version_link.pack()
        new_version_link.bind("<Button-1>", lambda e: self.callback("https://github.com/reinenichts/bookleter/releases"))

        button = tk.Button(
            toplevel, 
            text="Cool", 
            fg="black",
            command=toplevel.destroy
            )
        button.pack()

    def callback(self, url):
        webbrowser.open(url)


def gui_main():
    root = tk.Tk()
    root.minsize(300, 400)
    app = Application(master=root)
    root.mainloop()

if __name__ == "__main__":
    gui_main()






"""
    Credits
    This part of the application uses an Open Source component.
    This function was ported from a go script originally
    You can find the source code of the open source project along with license information below.
    i am grateful to this developer for his/her contribution to open source.

    Project: (a6-booklet-on-a4) https://github.com/ERnsTL/a6-booklet-on-a4
    License (GNU General Public License v3.0) https://www.gnu.org/licenses/gpl-3.0.en.html 
"""

def foop(pages_count):
    if pages_count == 0:
        raise ValueError("Enter number of pages")
    
    if pages_count % 8 != 0:
        raise ValueError("number of pages must be multiple of 8")

    sheets_count = pages_count // 8

    a4_sheets = [{"front": [None for _ in range(4)], "back": [None for _ in range(4)] } for i in range(sheets_count)]
    cur_page = 1

    for sheet_number in range(sheets_count):
        # front
        a4_sheets[sheet_number]["front"][0] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["front"][1] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["front"][2] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["front"][3] = cur_page
        cur_page += 1

        # back
        a4_sheets[sheet_number]["back"][0] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["back"][1] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["back"][2] = cur_page
        cur_page += 1
        a4_sheets[sheet_number]["back"][3] = cur_page
        cur_page += 1

    # vertically rip in half horizontally
    a5_sheets = [{"front": [None for _ in range(2)], "back": [None for _ in range(2)]} for i in range(sheets_count * 2)]

    # top half
    for sheet_number in range(sheets_count):
        # front
        a5_sheets[sheet_number]["front"][0] = a4_sheets[sheet_number]["front"][0]
        a5_sheets[sheet_number]["front"][1] = a4_sheets[sheet_number]["front"][1]

        # back
        a5_sheets[sheet_number]["back"][0] = a4_sheets[sheet_number]["back"][0]
        a5_sheets[sheet_number]["back"][1] = a4_sheets[sheet_number]["back"][1]

    # bottom half
    for sheet_number in range(sheets_count):
        # front
        a5_sheets[sheets_count + sheet_number]["front"][0] = a4_sheets[sheet_number]["front"][2]
        a5_sheets[sheets_count + sheet_number]["front"][1] = a4_sheets[sheet_number]["front"][3]

        # back
        a5_sheets[sheets_count + sheet_number]["back"][0] = a4_sheets[sheet_number]["back"][2]
        a5_sheets[sheets_count + sheet_number]["back"][1] = a4_sheets[sheet_number]["back"][3]


    # number all physical pages in book order
    cur_page = 0

    # left pages of the stack going up = first half book pages
    book_order = [None for _ in range(pages_count)]
    for sheet_number in range(len(a5_sheets)):
        # back right
        book_order[cur_page] = a5_sheets[sheet_number]["back"][1]
        cur_page += 1
        # front left
        book_order[cur_page] = a5_sheets[sheet_number]["front"][0]
        cur_page += 1

    for sheet_number in reversed(range(len(a5_sheets))):
        # front right
        book_order[cur_page] = a5_sheets[sheet_number]["front"][1]
        cur_page += 1
        # back left
        book_order[cur_page] = a5_sheets[sheet_number]["back"][0]
        cur_page += 1

    # apply translation
    print_order = [None for _ in range(pages_count)]
    for index, pg_number in enumerate(book_order):
        print_order[pg_number - 1] = index + 1

    return print_order


import sys, logging, re
from pathlib import Path, PurePath
from bookleter import Booklet
from bookleter import tkgui

def main():
    logging.basicConfig(level=logging.NOTSET)

    example_usage_command = """
        $ bookleter my_book.pdf 1-30 rtl '50 50 50 50'
        $ bookleter [pdfname] [start_page-end_page] [direction: rtl ltr] [page crop: '50 50 50 50']
        direction: right to left (rtl) or left to right (ltr)
        page crop: amount of page border crop in pixels 'left top right bottom'
    """

    if len(sys.argv) == 1:
        tkgui.gui_main()
    elif len(sys.argv) < 5:
        logging.error("missing some arguments\nuse it like this:\n{}".format(example_usage_command))
        sys.exit()
    else:
        start_page_number = int(sys.argv[2].split("-")[0])
        end_page_number = int(sys.argv[2].split("-")[1])

        book_direction = sys.argv[3]

        if not re.match("\\d+ \\d+ \\d+ \\d+", sys.argv[4]):
            logging.error("invalid crop option\nuse it like this:\n{}".format(example_usage_command))
            sys.exit()
        crop_amount = [int(s) for s in re.findall(r'\b\d+\b', sys.argv[4])]

        crop = {
            "left": crop_amount[0],
            "top": crop_amount[1],
            "right": crop_amount[2],
            "bottom": crop_amount[3]
        }

        try:
            NewBook = Booklet.Book(sys.argv[1], start_page_number, end_page_number, book_direction, crop)
            NewBook.make_booklet()
        except IndexError:
            print("Page number out of range")
        except Exception as e:
            raise e



if __name__ == '__main__':
    main()
